/*************************************************************************
* Copyright (c) 2008 High Energy Accelerator Reseach Organization (KEK)
*
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
**************************************************************************
* devMbboDirectF3RP61.c - Device Support Routines for F3RP61 multi-bit binary
* Output
*
*      Author: Jun-ichi Odagiri 
*      Date: 6-30-08
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "alarm.h"
#include "dbDefs.h"
#include "dbAccess.h"
#include "dbScan.h"
#include "recGbl.h"
#include "recSup.h"
#include "devSup.h"
#include "mbboDirectRecord.h"
#include "epicsExport.h"

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <asm/fam3rtos/m3iodrv.h>
#include "drvF3RP61.h"

extern int f3rp61_fd;

/* Create the dset for devMbboDirectF3RP61 */
static long init_record();
static long write_mbboDirect();

struct {
	long		number;
	DEVSUPFUN	report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	write_mbboDirect;
	DEVSUPFUN	special_linconv;
}devMbboDirectF3RP61={
	6,
	NULL,
	NULL,
	init_record,
	f3rp61GetIoIntInfo,
	write_mbboDirect,
	NULL
};
epicsExportAddress(dset,devMbboDirectF3RP61);

extern F3RP61_IO_INTR f3rp61_io_intr[M3IO_NUM_UNIT][M3IO_NUM_SLOT];

typedef struct {
  IOSCANPVT ioscanpvt; /* must comes first */
  union {
    M3IO_ACCESS_COM acom;
    M3IO_ACCESS_REG drly;
  } u;
  char device;
} F3RP61_LO_DPVT;


static long init_record(mbboDirectRecord *pmbboDirect)
{
  struct link *plink = &pmbboDirect->out;
  int size;
  char *buf;
  char *pC;
  F3RP61_LO_DPVT *dpvt;
  M3IO_ACCESS_COM *pacom;
  M3IO_ACCESS_REG *pdrly;
  int unitno, slotno, cpuno, start;
  char device;
  /*
  printf("devMbboDirectF3RP61: init_record() entered for %s\n", pmbboDirect->name);
  */
  /* bi.out must be an INST_IO */
  if (pmbboDirect->out.type != INST_IO) {
    recGblRecordError(S_db_badField,(void *)pmbboDirect,
		      "devMbboDirectF3RP61 (init_record) Illegal OUT field");
    pmbboDirect->pact = 1;
    return(S_db_badField);
  }

  size = strlen(plink->value.instio.string) + 1;

  buf = (char *) callocMustSucceed(size, sizeof(char), "calloc failed");
  strncpy(buf, plink->value.instio.string, size);
  buf[size - 1] = '\0';
  /*
  printf("buf:%s\n", buf);
  */
  pC = strchr(buf, ':');
  if (pC) {
    *pC++ = '\0';
    if (sscanf(pC, "U%d,S%d,X%d", &unitno, &slotno, &start) < 3) {
      errlogPrintf("devMbboDirectF3RP61: can't get interrupt source address for %s\n", pmbboDirect->name);
      pmbboDirect->pact = 1;
      return (-1);
    }
    if (f3rp61_register_io_interrupt(pmbboDirect, unitno, slotno, start) < 0) {
      errlogPrintf("devMbboDirectF3RP61: can't register I/O interrupt for %s\n", pmbboDirect->name);
      pmbboDirect->pact = 1;
      return (-1);
    }
  }
  if (sscanf(buf, "U%d,S%d,%c%d", &unitno, &slotno, &device, &start) < 4) {
    if (sscanf(buf, "CPU%d,R%d", &cpuno, &start) < 2) {
      errlogPrintf("devMbboDirectF3RP61: can't get I/O address for %s\n", pmbboDirect->name);
      pmbboDirect->pact = 1;
      return (-1);
    }
    else {
      device = 'R';
    }
  }
  if (!(device == 'Y' || device == 'A' || device == 'R')) {
    errlogPrintf("devMbboDirectF3RP61: illegal I/O address for %s\n", pmbboDirect->name);
    pmbboDirect->pact = 1;
    return (-1);
  }

  dpvt = (F3RP61_LO_DPVT *) callocMustSucceed(1,
					      sizeof(F3RP61_LO_DPVT),
					      "calloc failed");
  dpvt->device = device;
  if (device == 'R') {
    pacom = &dpvt->u.acom;
    pacom->cpuno = (unsigned short) cpuno;
    pacom->start = (unsigned short) start;
  } else {
    pdrly = &dpvt->u.drly;
    pdrly->unitno = (unsigned short) unitno;
    pdrly->slotno = (unsigned short) slotno;
    pdrly->start = (unsigned short) start;
  }
  switch (device) {
  case 'R':
    pacom->count = (unsigned short) 1;
    break;
  case 'Y':
  default:
    pdrly->count = (unsigned short) 1;
  }

  pmbboDirect->dpvt = dpvt;

  return(0);
}

static long write_mbboDirect(mbboDirectRecord *pmbboDirect)
{
  F3RP61_LO_DPVT *dpvt = (F3RP61_LO_DPVT *) pmbboDirect->dpvt;
  M3IO_ACCESS_COM *pacom = &dpvt->u.acom;
  M3IO_ACCESS_REG *pdrly = &dpvt->u.drly;
  char device = dpvt->device;
  void *p = (void *) pdrly;
  unsigned short wdata;
  int command;
  /*
  printf("devMbboDirectF3RP61: write_mbboDirect() entered for %s\n", pmbboDirect->name);
  */
  switch (device) {
  case 'Y':
    /*
    printf("devMbboDirectF3RP61: write_mbboDirect() for %s, with %d\n", pmbboDirect->name, pmbboDirect->val);
    */
    command = M3IO_WRITE_OUTRELAY;
    pdrly->u.outrly[0].data = (unsigned short) pmbboDirect->rval;
    pdrly->u.outrly[0].mask = (unsigned short) 0xffff;
    break;
  case 'R':
    command = M3IO_WRITE_COM;
    wdata = (unsigned short) pmbboDirect->rval;
    pacom->pdata = &wdata;
    p = (void *) pacom;
    break;
  default:
    command = M3IO_WRITE_REG;
    wdata = (unsigned short) pmbboDirect->rval;
    pdrly->u.pwdata = &wdata;
  }

  if (ioctl(f3rp61_fd, command, p) < 0) {
    errlogPrintf("devMbboDirectF3RP61: ioctl failed [%d] for %s\n", errno, pmbboDirect->name);
    return (-1);
  }
  pmbboDirect->udf=FALSE;

  return (0);
}
